/* =====================================================================================================================
 * File - main.cpp
 * ---------------------------------------------------------------------------------------------------------------------
 */

/* =====================================================================================================================
 * ---------------------------------------------------------------------------------------------------------------------
 */

#include <GLUT/glut.h>

#include <fstream>
#include <iostream>
#include <string>


/* =====================================================================================================================
 * ---------------------------------------------------------------------------------------------------------------------
 */

const int k_dxWINDOW	= 640;
const int k_dyWINDOW	= 480;


/* =====================================================================================================================
 * ---------------------------------------------------------------------------------------------------------------------
 */

int g_dxWindow;
int g_dyWindow;


/* =====================================================================================================================
 * ---------------------------------------------------------------------------------------------------------------------
 */

GLuint	g_vertex_shader;
GLuint	g_fragment_shader;
GLuint	g_geometry_shader;
GLuint	g_program;


/* =====================================================================================================================
 * ---------------------------------------------------------------------------------------------------------------------
 */

void RetrieveShaderLog(std::ostream& os, GLuint obj)
{
	int cbInfoLog = 0;
	glGetShaderiv(obj, GL_INFO_LOG_LENGTH, &cbInfoLog);
	
	if ( cbInfoLog > 0 )
	{
		char*	pbuffInfoLog = new char[cbInfoLog];
		int		cbInfoRead  = 0;
		glGetShaderInfoLog(obj, cbInfoLog, &cbInfoRead, pbuffInfoLog);

		os << "ShaderInfoLog: " << pbuffInfoLog << "\n";

		delete [] pbuffInfoLog;
	}
	else
	{
		printf("Shader Info Log: OK\n");
	}
}


/* =====================================================================================================================
 * ---------------------------------------------------------------------------------------------------------------------
 */

void RetrieveProgramLog(std::ostream& os, GLuint obj)
{
	int		cbInfoLog = 0;
	glGetProgramiv(obj, GL_INFO_LOG_LENGTH, &cbInfoLog);
	
	if ( cbInfoLog > 0 )
	{
		char*	pbuffInfoLog = new char[cbInfoLog];
		int		cbInfoRead  = 0;
		glGetProgramInfoLog(obj, cbInfoLog, &cbInfoRead, pbuffInfoLog);

		os << "ProgramLog: " << pbuffInfoLog << "\n";

		delete [] pbuffInfoLog;
	}
	else
	{
		os << "Program Info Log: OK\n";
	}
}


/* =====================================================================================================================
 * ---------------------------------------------------------------------------------------------------------------------
 */

std::string GetShader(const char* pszFilename)
{
	using std::getline;
	using std::ifstream;
	using std::string;
	
	string		str, strFile;
	ifstream	ifs(pszFilename);
	if ( ifs.is_open() )
	{
		while ( getline(ifs, str) )
		{
			strFile += str;
		}
	}
	
	return strFile;
}


/* =====================================================================================================================
 * ---------------------------------------------------------------------------------------------------------------------
 */

void ReshapeHandler(int w, int h)
{
	//stores the width and height
	g_dxWindow = w;
	g_dyWindow = h;
	glViewport(0, 0, g_dxWindow, g_dyWindow);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	gluOrtho2D(0, g_dxWindow, 0, g_dyWindow);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}


/* =====================================================================================================================
 * ---------------------------------------------------------------------------------------------------------------------
 */

void DisplayHandler(void)
{
	glClearColor(0.0, 0.0, 0.0, 1.0);

	glClear(GL_COLOR_BUFFER_BIT);

	glViewport(0, 0, g_dxWindow, g_dyWindow);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	gluOrtho2D(0, g_dxWindow, 0, g_dyWindow);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();


	// sorry, to lazy to add more interesting geometry

	glBegin(GL_LINES);

		glVertex2i(g_dxWindow / 2, 0);
		glVertex2i(g_dxWindow / 2, g_dyWindow);

	glEnd();

	glutSwapBuffers();
}


/* =====================================================================================================================
 * ---------------------------------------------------------------------------------------------------------------------
 */

void KeyboardHandler(unsigned char key, int x, int y)
{
	switch ( key )
	{
		case 27:
		case 'q':
		case 'Q':
			exit(0);
	}
}


/* =====================================================================================================================
 * ---------------------------------------------------------------------------------------------------------------------
 */

void SetupShaders()
{
	using std::cout;
	using std::string;
	
	g_vertex_shader		= glCreateShader(GL_VERTEX_SHADER);
	g_fragment_shader	= glCreateShader(GL_FRAGMENT_SHADER);
	g_geometry_shader	= glCreateShader(GL_GEOMETRY_SHADER_EXT);

	string	vs			= GetShader("shader.vert");
	const char*	pvs		= vs.c_str();
	glShaderSource(g_vertex_shader, 1, &pvs, NULL);
	glCompileShader(g_vertex_shader);

	string	fs			= GetShader("shader.frag");
	const char*	pfs		= fs.c_str();
	glShaderSource(g_fragment_shader, 1, &pfs, NULL);
	glCompileShader(g_fragment_shader);

	g_program = glCreateProgram();

	glAttachShader(g_program, g_fragment_shader);
	glAttachShader(g_program, g_vertex_shader);

	glProgramParameteriEXT(g_program, GL_GEOMETRY_INPUT_TYPE_EXT, GL_LINES);
	glProgramParameteriEXT(g_program, GL_GEOMETRY_OUTPUT_TYPE_EXT, GL_LINE_STRIP);

	int		max_vertices;
	glGetIntegerv(GL_MAX_GEOMETRY_OUTPUT_VERTICES_EXT, &max_vertices);
	glProgramParameteriEXT(g_program, GL_GEOMETRY_VERTICES_OUT_EXT, max_vertices);

	glLinkProgram(g_program);
	glUseProgram(g_program);

	RetrieveShaderLog(cout, g_vertex_shader);
	RetrieveShaderLog(cout, g_fragment_shader);
	RetrieveProgramLog(cout, g_program);
}


/* =====================================================================================================================
 * ---------------------------------------------------------------------------------------------------------------------
 */

int main(int argc, char** argv)
{
	glutInit(&argc, argv);

	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA);

	glutInitWindowPosition(10, 50);
	glutInitWindowSize(k_dxWINDOW, k_dyWINDOW);
	glutCreateWindow("OpenGL Shader Language Demo");

	glutDisplayFunc(DisplayHandler);
	glutIdleFunc(DisplayHandler);
	glutReshapeFunc(ReshapeHandler);
	glutKeyboardFunc(KeyboardHandler);

	glClearColor(1.0, 1.0, 1.0, 1.0);

	SetupShaders();

	glutMainLoop();

	return 0;
}
